import "moebius92.CCTimer.CCTimerOptionsButton";
import "moebius92.CCTimer.CCTimerOptionsCCEntry";
import "moebius92.CCTimer.CCTimerOptionsCCLabel";
import "moebius92.CCTimer.CCTimerOptionsControl";
import "moebius92.CCTimer.CCTimerOptionsEffect";
import "moebius92.CCTimer.CCTimerOptionsEffectHeader";
import "moebius92.CCTimer.CCTimerOptionsFixedField";
import "moebius92.CCTimer.CCTimerOptionsLabeledField";
import "moebius92.CCTimer.CCTimerOptionsQuickslot";

local function isnonneg(x)
	x = tonumber(x);
	if x == nil or x < 0 then
		return false;
	end
	return x;
end

local function isnumber(x)
	x = tonumber(x);
	if x == nil then
		return false;
	end
	return x;
end

local function nilnumber(x) 
	if x ~= "?" then
		x = tonumber(x);
		if x == nil then
			return false;
		end
		return x;
	end
	return nil;
end

local function tonilstring(x)
	if x == nil then
		return "?"
	end
	return tostring(x);
end

local function zeropercent(x)
	if x ~= "" then
		x = tonumber(x);
		if x == nil then
			return false;
		end
		if x == 0 then
			return nil;
		end
		return x / 100;
	end
	return nil;
end

local function nilpercent(x)
	if x ~= "?" then
		x = tonumber(x);
		if x == nil then
			return false;
		end
		return x / 100;
	end
	return nil;
end

local function percent(x)
	x = tonumber(x);
	if x == nil then
		return false;
	end
	return x / 100;
end

local function topercent(x)
	if x == nil then
		return "?"
	end
	return tostring(x * 100);
end

local function tozpercent(x)
	if x == nil then
		return ""
	end
	return tostring(x * 100);
end

local function tableCopy(t)
	if type(t) ~= "table" then
		return t;
	end

	local r = { };
	for k, v in pairs(t) do
		r[k] = tableCopy(v);
	end
	return r;
end

local function fnMouseClick(self, args, intersection)
	if self.window.ccSkills[self.shortcut].regimes.options:IsVisible() then
		self:SetBrightness(0.25);
		self.window.ccSkills[self.shortcut].regimes.options:SetVisible(false);
	else
		self:SetBrightness(0.50);
		self.window.ccSkills[self.shortcut].regimes.options:SetVisible(true);
	end
end

local function fnMouseEnter(self, args, intersection)
	if intersection ~= nil then
		intersection:SetBrightness(2.0);
		local t, b = intersection:GetBounds();
		for shortcut, regime in pairs(self.options.highlightRegimes) do
			if shortcut ~= self.shortcut then
				regime:SetBounds(t, b);
				regime:SetVisible(true);
			end
		end
	end
end

local function fnMouseLeave(self, args, intersection)
	if intersection ~= nil then
		intersection:SetBrightness(1.0);
		for _, regime in pairs(self.options.highlightRegimes) do
			regime:SetVisible(false);
		end
	end
end

CCTimerOptions = class(Turbine.UI.Control);

function CCTimerOptions:Constructor(window)
	Turbine.UI.Control.Constructor(self);

	self.window = window;
	self.width = 65;
	self.height = 20;

	self:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.3));

	self.selectButtons = { };
	table.insert(self.selectButtons, CCTimerOptionsControl(self, "Skills", 75, self.height, function() self:SetState(1); end));
	table.insert(self.selectButtons, CCTimerOptionsControl(self, "Effects", 75, self.height, function() self:SetState(2); end));

	self.effectsPanel = { };
	self.effectsPanel.header = { };
	table.insert(self.effectsPanel.header, CCTimerOptionsEffectHeader(self, self.width, self.height, { "Att", "Ind" }));
	self.effectsPanel.effects = { };
	self:InitEffects();

	self.effectsPanel.clean = { };
	table.insert(self.effectsPanel.clean, CCTimerOptionsButton(self, "Clean", function(self, args)
		self.parent:InitEffects();
		self.parent:Redraw();
	end));
	table.insert(self.effectsPanel.clean, CCTimerOptionsButton(self, "Default", function(self, args)
		self.parent:AddEffects(self.parent.window:AddDefaultTimingEffects());
		self.parent:Redraw();
	end));

	self.configPanel = { };
	self.configPanel.calibrate = { };
	table.insert(self.configPanel.calibrate, CCTimerOptionsFixedField(self, "Apply", self.width, 40, self.height, function(x) if x == nil then return ""; end return string.format("%.3f", x); end));
	table.insert(self.configPanel.calibrate, CCTimerOptionsFixedField(self, "Key", self.width, 40, self.height, function(x) if x == nil then return ""; end return string.format("%.3f", x); end));
	table.insert(self.configPanel.calibrate, CCTimerOptionsControl(self, "Calibrate", 150, self.height, function() self.window:ToggleCalibrate(); end));
	self.configPanel.calibrate[3]:SetColor(0.5);

	self.configPanel.cc = { };

	self.highlightRegimes = { };

	window.regimeContainers.options = CCTimerRegimeContainer(window.colors);
	window.regimeContainers.options.shortcut = "options";
	window.regimeContainers.options.options = self;
	window.regimeContainers.options.MouseEnter = fnMouseEnter;
	window.regimeContainers.options.MouseLeave = fnMouseLeave;
	window.regimeContainers.options:SetBrightness(0.50);
	window.regimeContainers.options:SetSize(window.size + 4, window:GetHeight());
	for _, skill in pairs(window.ccSkills) do
		skill.regimes.options = window.regimeContainers.options:AddRegime();
	end
	window.regimesRecovery.options = window.regimeContainers.options:AddRegime();
	window.regimesRecovery.options:SetBounds(window:TimeY(-0.01), window:TimeY(-1.0));
	window.regimesRecovery.options:SetVisible(true);
	self.highlightRegimes.options = window.regimeContainers.options:AddRegime();
	self.highlightRegimes.options:SetBrightness(2.0);
	table.insert(self.configPanel.cc, CCTimerOptionsCCLabel(self, window.regimeContainers.options, self.width, self.height, window.size, 5, { "Duration", "Pre-root", "Apply", "Start", "Stop" }, window.colors));

	local shortcuts = { };
	for shortcut in pairs(window.ccSkills) do
		table.insert(shortcuts, shortcut);
	end
	for _, shortcut in ipairs(shortcuts) do
		window.regimeContainers[shortcut] = CCTimerRegimeContainer(window.colors);
		window.regimeContainers[shortcut].shortcut = shortcut;
		window.regimeContainers[shortcut].window = window;
		window.regimeContainers[shortcut].options = self;
		window.regimeContainers[shortcut].MouseClick = fnMouseClick;
		window.regimeContainers[shortcut].MouseEnter = fnMouseEnter;
		window.regimeContainers[shortcut].MouseLeave = fnMouseLeave;
		window.regimeContainers[shortcut]:SetBrightness(0.50);
		window.regimeContainers[shortcut]:SetSize(window.size + 4, window:GetHeight());
		window.ccSkills[shortcut].regimes.entry = window.regimeContainers[shortcut]:AddRegime();
		window.ccSkills[shortcut].regimes.entry:SetVisible(true);
		self.highlightRegimes[shortcut] = window.regimeContainers[shortcut]:AddRegime();
		self.highlightRegimes[shortcut]:SetBrightness(2.0);
		if window.config[window.curConfig][shortcut].preroot ~= nil then
			table.insert(self.configPanel.cc, CCTimerOptionsCCEntry(self, window.regimeContainers[shortcut], self.width, self.height, window.size, 5, shortcut, window.colors,
			                                                        { "duration", "preroot", "apply", "start", "stop" },
			                                                        { isnonneg, isnonneg, isnumber, nilnumber, nilnumber },
			                                                        { tostring, tostring, tostring, tonilstring, tonilstring }));
		else
			table.insert(self.configPanel.cc, CCTimerOptionsCCEntry(self, window.regimeContainers[shortcut], self.width, self.height, window.size, 5, shortcut, window.colors,
			                                                        { "duration", "", "apply", "start", "stop" },
			                                                        { isnonneg, isnonneg, isnumber, nilnumber, nilnumber },
			                                                        { tostring, tostring, tostring, tonilstring, tonilstring }));
		end
	end

	self.configPanel.induction = { };
	table.insert(self.configPanel.induction, CCTimerOptionsLabeledField(self, self.width * 2, self.width, self.height, self.height, "Induction", "%", nil, nil, "induction", nilpercent, topercent));
	table.insert(self.configPanel.induction, CCTimerOptionsLabeledField(self, self.width * 2, self.width, self.height, self.height, " + Preview", "%", self, nil, "induction", percent, topercent));

	self.configPanel.attdur = { };
	table.insert(self.configPanel.attdur, CCTimerOptionsLabeledField(self, self.width * 2, self.width, self.height, self.height, "Attack Duration", "%", nil, nil, "attdur", nilpercent, topercent));
	table.insert(self.configPanel.attdur, CCTimerOptionsLabeledField(self, self.width * 2, self.width, self.height, self.height, " + Preview", "%", self, nil, "attdur", percent, topercent));

	self.configPanel.skill = { };
	table.insert(self.configPanel.skill, CCTimerOptionsQuickslot(self, window.size));

	self.setupName = Turbine.UI.Lotro.TextBox();
	self.setupName:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self.setupName.FocusGained = function(self, args)
		self.value = self:GetText();
	end;
	self.setupName.FocusLost = function(self, args)
		self:SetText(self.value);
		self.value = nil;
	end;
	self.setupName.KeyDown = function(self, args)
		if args.Action == 162 then
			local config = self:GetText();

			if config ~= self.parent.window.curConfig then
				self.parent.window.config[config] = self.parent.window.config[self.parent.window.curConfig];
				self.parent.window.config[self.parent.window.curConfig] = nil;
				self.parent.window.curConfig = config;
				self.parent.window:SaveSettings();
			end
			self.value = config;
			self.parent:Focus();
		end
	end;
	self.setupName:SetMultiline(false);
	self.setupName:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.setupName:SetParent(self);
	self.setupName:SetZOrder(0);
	self.setupName.parent = self;
	self.setupMenu = Turbine.UI.ContextMenu();
	self.setupDropDown = Turbine.UI.Button();
	self.setupDropDown:SetBackground("moebius92/CCTimer/res/dropdown.tga");
	self.setupDropDown:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.setupDropDown:SetSize(self.height - 4, self.height);
	self.setupDropDown:SetZOrder(1);
	self.setupDropDown:SetParent(self);
	self.setupDropDown.parent = self;
	self.setupDropDown.MouseClick = function(self, args)
		local items = self.parent.setupMenu:GetItems();
		local f = function(self, args) local config = self:GetText(); if config ~= self.parent.window.curConfig then self.parent.window:LoadConfig(config); end end;
		local configs = { };

		items:Clear();
		for label in pairs(self.parent.window.config) do
			table.insert(configs, label);
		end
		table.sort(configs);
		for _, label in ipairs(configs) do
			local item = Turbine.UI.MenuItem(label);
			item.parent = self.parent;
			item.Click = f;
			items:Add(item);
		end
		self.parent.setupMenu:ShowMenuAt(self.parent.setupName:PointToScreen(0, self.parent.setupName:GetHeight()));
	end;

	self.configPanel.config = { };
	table.insert(self.configPanel.config, CCTimerOptionsButton(self, "Copy", function(self, args)
		local curConfig = self.parent.window.curConfig;
		local name = curConfig .. " (copy)";

		if self.parent.window.config[name] ~= nil then
			local i = 1;
			while true do
				name = curConfig .. " (copy " .. tostring(i) .. ")";
				if self.parent.window.config[name] == nil then
					break;
				end
				i = i + 1;
			end
		end

		self.parent.window.config[name] = tableCopy(self.parent.window.config[curConfig]);
		self.parent.window:SaveSettings();
		self.parent.window:LoadConfig(name);
	end));
	table.insert(self.configPanel.config, CCTimerOptionsButton(self, "Delete", function(self, args)
		local n = 0;

		for k in pairs(self.parent.window.config) do
			n = n + 1;
		end

		if n > 1 then
			self.parent.window.config[self.parent.window.curConfig] = nil;
			for k in pairs(self.parent.window.config) do
				self.parent.window:LoadConfig(k);
				break;
			end
			self.parent.window:SaveSettings();
		end
	end));

	self:UpdateEntries();
	self:SetState(1);
end

function CCTimerOptions:InitEffects()
	for _, effect in ipairs(self.effectsPanel.effects) do
		effect:SetParent(nil);
		effect:SetVisible(false);
	end
	self.effectsPanel.effects = { };

	for iconName, timing in pairs(self.window.timingEffects) do
		local i = string.find(iconName, ":");

		if i ~= nil then
			local icon = tonumber(string.sub(iconName, 1, i - 1), 16);
			local name = string.sub(iconName, i + 1)
			
			table.insert(self.effectsPanel.effects, CCTimerOptionsEffect(self, self.width, self.height, icon, name, { "ad", "ind" }, { timing.ad, timing.ind }, zeropercent, tozpercent));
		end
	end

	table.sort(self.effectsPanel.effects, function(a, b) return a.name < b.name; end);
end

function CCTimerOptions:LogEffect(icon, name)
	for _, effect in ipairs(self.effectsPanel.effects) do
		if effect.icon == icon and effect.name == name then
			return;
		end
	end

	table.insert(self.effectsPanel.effects, CCTimerOptionsEffect(self, self.width, self.height, icon, name, { "ad", "ind" }, { }, zeropercent, tozpercent));
	table.sort(self.effectsPanel.effects, function(a, b) return a.name < b.name; end);
	self:Redraw();
end

function CCTimerOptions:AddEffects(added)
	for _, iconName in ipairs(added) do
		local timing = self.window.timingEffects[iconName];
		if timing ~= nil then
			local i = string.find(iconName, ":");

			if i ~= nil then
				local icon = tonumber(string.sub(iconName, 1, i - 1), 16);
				local name = string.sub(iconName, i + 1)

				for j, effect in ipairs(self.effectsPanel.effects) do
					if effect.icon == icon and effect.name == name then
						table.remove(self.effectsPanel.effects, j);
						break;
					end
				end
				table.insert(self.effectsPanel.effects, CCTimerOptionsEffect(self, self.width, self.height, icon, name, { "ad", "ind" }, { timing.ad, timing.ind }, zeropercent, tozpercent));
			end
		end
	end
	table.sort(self.effectsPanel.effects, function(a, b) return a.name < b.name; end);
end

function CCTimerOptions:UpdateEntries()
	local config = self.window.config[self.window.curConfig];

	self.setupName:SetText(self.window.curConfig);

	for i, entry in ipairs(self.configPanel.cc) do
		if i == 1 then
			for field, color in pairs(entry.colors) do
				color:SetBackColor(self.window.colors[config[field].color]);
			end
		else
			for _, field in pairs(entry.fields) do
				entry:SetValue(field.entry, self.window.config[self.window.curConfig][entry.shortcut][field.entry]);
			end
			entry.color:SetBackColor(self.window.colors[config[entry.shortcut].color]);

			if self.window.ccSkills[entry.shortcut].regimes.options:IsVisible() then
				self.window.regimeContainers[entry.shortcut]:SetBrightness(0.50);
			else
				self.window.regimeContainers[entry.shortcut]:SetBrightness(0.25);
			end
		end
	end

	self.induction = 0;
	self.attdur = 0;
	self.configPanel.induction[1]:SetValue(self.window.config[self.window.curConfig].induction);
	self.configPanel.induction[2]:SetValue(self.induction);
	self.configPanel.attdur[1]:SetValue(self.window.config[self.window.curConfig].attdur);
	self.configPanel.attdur[2]:SetValue(self.attdur);

	if config.skill == nil then
		self.configPanel.skill[1]:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, ""));
	else
		self.configPanel.skill[1]:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, config.skill));
	end
end

function CCTimerOptions:SetShortcut(value)
	self.window.config[self.window.curConfig].skill = value;
	self.window:SaveSettings();
end

function CCTimerOptions:SetSetting(root, shortcut, field, value)
	if root == nil then
		root = self.window.config[self.window.curConfig];
	end
	if shortcut ~= nil then
		root = root[shortcut];
	end
	root[field] = value;
	self.window:SaveSettings();
	self.window:UpdateRegimes();
end

function CCTimerOptions:SetEffect(icon, name, label, value)
	local iconName = string.format("%8X", icon) .. ":" .. name;

	if value == nil then
		if self.window.timingEffects[iconName] ~= nil then
			self.window.timingEffects[iconName][label] = nil;
			if next(self.window.timingEffects[iconName]) == nil then
				self.window.timingEffects[iconName] = nil;
			end
		end
	else
		if self.window.timingEffects[iconName] == nil then
			self.window.timingEffects[iconName] = { };
		end
		self.window.timingEffects[iconName][label] = value;
	end
	self.window:SaveSettings();
	self.window:UpdateRegimes();
end

function CCTimerOptions:SetColorSetting(shortcut, color, save)
	self.window.config[self.window.curConfig][shortcut].color = color;
	if save then
		self.window:SaveSettings();
	end
	self.window:UpdateRegimes();
end

function CCTimerOptions:GetColorSetting(shortcut)
	return self.window.config[self.window.curConfig][shortcut].color;
end

function CCTimerOptions:Calibrate(apply, key)
	self.configPanel.calibrate[1]:SetValue(apply);
	self.configPanel.calibrate[2]:SetValue(key);
end

function CCTimerOptions:StartCalibrate()
	self.configPanel.calibrate[3]:SetColor(1.0);
end

function CCTimerOptions:StopCalibrate(apply, key)
	self.configPanel.calibrate[3]:SetColor(0.5);
end

function CCTimerOptions:SetState(state)
	if self.state ~= state then
		self.state = state;
		for state, button in ipairs(self.selectButtons) do
			if state == self.state then
				button:SetColor(1.0);
			else
				button:SetColor(0.5);
			end
		end
		self:Redraw();
	end
end

function CCTimerOptions:Layout(controls, y, h, m)
	local w = self:GetWidth();
	local widths = { };
	local q = { };
	local c, i, l, p, x;

	for i, control in ipairs(controls) do
		widths[i] = control:GetWidth();
	end

	i = 1;
	c = 0;
	p = { };
	while widths[i] ~= nil do
		l = w - 4 * m;
		l = l - widths[i];
		repeat
			table.insert(q, { i = i + 1, c = c + (l * l), p = { e = i, p = p } });
			i = i + 1;
			if widths[i] ~= nil then
				l = l - (m + widths[i]);
			else
				break;
			end
		until l <= 0;
		table.sort(q, function(a, b) return a.c < b.c; end);
		i = table.remove(q, 1);
		c = i.c;
		p = i.p;
		i = i.i;
	end

	q = { };
	while p.p ~= nil do
		table.insert(q, 1, p.e);
		p = p.p;
	end

	c = 1;
	for _, v in ipairs(q) do
		x = w;
		for i = c, v, 1 do
			x = x - widths[i];
		end
		x = (x - (v - c) * m) / 2;
		for i = c, v, 1 do
			controls[i]:SetPosition(x, y);
			x = x + widths[i] + m;
		end

		y = y + h + m;
		c = v + 1;
	end
	return y;
end

function CCTimerOptions:SizeChanged()
	local w, h = self:GetSize();

	if self.w ~= w then
		self:Redraw();
	end
end

function CCTimerOptions:Redraw()
	local w, h = self:GetSize();

	local y = 10;

	y = self:Layout(self.selectButtons, y, self.height, 10) + 10;

	for _, row in pairs(self.configPanel) do
		for _, item in ipairs(row) do
			item:SetVisible(self.state == 1);
		end
	end

	for name, row in pairs(self.effectsPanel) do
		for _, item in ipairs(row) do
			if self.state == 2 then
				item:SetVisible(true);
				if name ~= "clean" then
					item:SetWidth(w);
				end
			else
				item:SetVisible(false);
			end
		end
	end

	if self.state == 1 then
		y = self:Layout(self.configPanel.calibrate, y, self.height, 10) + 10;

		y = self:Layout(self.configPanel.cc, y, self.configPanel.cc[1]:GetHeight(), 15) + 10;

		y = self:Layout(self.configPanel.induction, y, self.height, 2);
		y = self:Layout(self.configPanel.attdur, y, self.height, 2) + 10;

		y = self:Layout(self.configPanel.skill, y, self.window.size, 10) + 10;

		local l = math.min(w - 4, self.width * 3.5);
		x = (w - l) / 2;
		self.setupName:SetSize(l, self.height);
		self.setupName:SetPosition(x, y);
		x = x + (l - (self.height - 1));
		self.setupDropDown:SetPosition(x, y);
		y = y + self.height + 5;

		y = self:Layout(self.configPanel.config, y, self.window.size, 5) + 5;
	elseif self.state == 2 then
		self.effectsPanel.header[1]:SetPosition(0, y);
		y = y + self.height + 5;
		for _, effect in ipairs(self.effectsPanel.effects) do
			effect:SetPosition(0, y);
			y = y + self.height + 5;
		end
		y = y + 10;
		y = self:Layout(self.effectsPanel.clean, y, self.window.size, 5) + 5;
	end

	self:SetHeight(y);
	self.w = w;
end