CCTimerOptionsColorSquare = class(Turbine.UI.Control);

function CCTimerOptionsColorSquare:Constructor(parent)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self:SetParent(parent);
	self:SetMouseVisible(true);
end

function CCTimerOptionsColorSquare:MouseEnter(args)
	self.parent:EnterIndex(self.index);
end

function CCTimerOptionsColorSquare:MouseLeave(args)
	self.parent:LeaveIndex(self.index);
end

function CCTimerOptionsColorSquare:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.parent:ClickIndex(self.index);
	end
end

CCTimerOptionsColor = class(Turbine.UI.Window);

function CCTimerOptionsColor:Constructor()
	Turbine.UI.Window.Constructor(self);

	self.target = nil;
	self.colors = { };
	self:SetMouseVisible(true);
end

function CCTimerOptionsColor:SetTarget(target)
	local m = #target.colors;

	self.target = target;

	for i = #self.colors, m, 1 do
		table.insert(self.colors, CCTimerOptionsColorSquare(self));
	end

	local y = 0;
	local index = (target.options:GetColorSetting(target.shortcut) - math.floor((m - 1) / 2)) % m;
	if index == 0 then
		index = m;
	end
	for i, color in ipairs(self.colors) do
		if i <= m then
			color:SetSize(target.size, target.size);
			color:SetPosition(0, y);
			color:SetBackColor(target.colors[index]);
			color:SetVisible(true);
			color.index = index;

			y = y + target.size;
			index = index + 1;
			if index > m then
				index = index - m;
			end
		else
			color:SetVisible(false);
		end
	end

	self:SetSize(target.size, y);
	local x, y = target:PointToScreen(0, 0);
	self:SetPosition(x, y - math.floor((m - 1) / 2) * target.size);
end

function CCTimerOptionsColor:FocusGained(args)
	self.colorI = self.target.options:GetColorSetting(self.target.shortcut);
end

function CCTimerOptionsColor:FocusLost(args)
	self:SetVisible(false);
end

function CCTimerOptionsColor:EnterIndex(index)
	self.target.options:SetColorSetting(self.target.shortcut, index, false);
end

function CCTimerOptionsColor:LeaveIndex(index)
	self.target.options:SetColorSetting(self.target.shortcut, self.colorI, false);
end

function CCTimerOptionsColor:ClickIndex(index)
	self.colorI = index;
	self.target:SetBackColor(self.target.colors[index]);
	self.target.options:SetColorSetting(self.target.shortcut, index, true);
	self:SetVisible(false);
end

CCTimerOptionsColorSelect = class(Turbine.UI.Control);

CCTimerOptionsColorSelect.color = CCTimerOptionsColor();

function CCTimerOptionsColorSelect:Constructor(parent, options, shortcut, colors, size)
	Turbine.UI.Control.Constructor(self);

	self.options = options;
	self.shortcut = shortcut;
	self.colors = colors;
	self.size = size;

	self:SetParent(parent);
	self:SetSize(size, size);
end

function CCTimerOptionsColorSelect:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		CCTimerOptionsColorSelect.target = self;
		self.color:SetTarget(self);
		self.color:SetVisible(true);
		self.color:Activate();
		self.color:Focus();
	end
end