import "moebius92.ActiveSkills";
import "moebius92.Handlers";
import "moebius92.RaceTrack";
import "moebius92.CCTimer.CCTimerRegime";
import "moebius92.CCTimer.CCTimerOptions";

-- Codes for the keys used to activate quickslots 1 through 72
local quickKeys = { 0x1000002A, 0x10000032, 0x1000003E, 0x10000047, 0x10000050, 0x10000057,
                    0x1000005F, 0x10000067, 0x10000071, 0x1000004F, 0x10000056, 0x1000005E,

                    0x10000066, 0x10000070, 0x10000078, 0x10000082, 0x1000008E, 0x10000096,
                    0x1000009E, 0x10000011, 0x10000019, 0x10000022, 0x10000029, 0x10000031,

                    0x1000003D, 0x10000046, 0x1000004E, 0x10000055, 0x1000005D, 0x10000005,
                    0x1000000B, 0x10000010, 0x10000018, 0x10000021, 0x10000028, 0x10000030,

                    0x1000003C, 0x10000045, 0x1000004D, 0x1000008D, 0x10000095, 0x1000009D,
                    0x100000A3, 0x100000AD, 0x100000B0, 0x100000B9, 0x10000004, 0x1000000A,

                    0x1000000F, 0x10000020, 0x10000027, 0x1000002F, 0x1000003B, 0x10000044,
                    0x1000004C, 0x10000054, 0x1000005C, 0x10000065, 0x1000006F, 0x100000AC,

                    0x100000AF, 0x100000B8, 0x10000003, 0x10000009, 0x1000000E, 0x10000017,
                    0x1000001F, 0x10000026, 0x1000002E, 0x1000009C, 0x100000A2, 0x100000AB };

-- Tracks the remaining time of crowd control effects.  Also display the delay between a skill being clicked on and the resulting CC effect being applied.
-- This can be used to determine if it is "safe" to use a CC skill.  For example, if Blinding Flash applies Dazed 1.3 seconds after Blinding Flassh is clicked,
-- then if Blinding Flash is used on a Dazed or Stunned targetd with between 1.3 and 0.3 seconds, then the Blinding Flash will not apply Dazed, since the target
-- will have the Recovering effect (post-Dazed/Stunned effect).  The main things to watch out for are Recovery (post-Dazed/Stunned - no CC effect can overwrite it)
-- and Rooted (Rooted targets can not be re-Rooted - you have to wait for Rooted to end, first)
CCTimerWindow = class(Turbine.UI.Window);

CCTimerWindow.keymap = { };
for quickslot, key in pairs(quickKeys) do
	CCTimerWindow.keymap[key] = quickslot;
end

CCTimerWindow.ccEffects = {
	["410D8360:Dazed"]						= "0x70003152",	-- Blinding Flash
	["410D7BA0:Feared"]						= "0x70003E82",	-- Invocation of Elbereth
	["4100823F:Knocked Down"]				= "0x70003EC2",	-- Light of the Rising Dawn
	["4100823F:Knocked Out"]				= "0x70003EC2",	-- Light of the Rising Dawn
	["410E7686:Recovering"]					= 0,
	["4100823D:Rooted"]						= "0x70003EC0",	-- Herb-lore
	["4100823F:Stunned"]					= "0x70003EC2",	-- Light of the Rising Dawn
	["41008214:Vulnerable"]					= "0x70003F0D"	-- Exploit Opening
};

CCTimerWindow.otherEffects = {
	["410056A2:The Shaking of Arda"]		= "0x70003143",	-- Cracked Earth
	["410DCD82:Permanent State Immunity"]	= "0x70003EBC",	-- Sign of Power: Righteousness
	["4100823C:Temporary State Immunity"]	= "0x70003EBC",	-- Sign of Power: Righteousness
	["410DCD82:Temporary State Immunity"]	= "0x70003EBC"	-- Sign of Power: Righteousness
};

CCTimerWindow.defaultTimingEffects = {
	["410EA251:Adroit Manoeuvre - Induction Reduction"]	= {             ind = -0.20 },
	["410087C5:Anthem of Prowess"]						= { ad = -0.05              },
	["41110AC8:Coordination"]							= {             ind = -1.00 },
	["4100C2AC:Distracting Cloud"]						= { ad =  0.20              },
	["410036A1:Entangled"]								= { ad =  0.25              },
	["41139EF1:Feral - Feline"]							= {             ind = -0.10 },
	["410E99FD:Grasp of Darkness"]						= { ad =  0.50, ind =  0.50 },
--	["410D71F3:The Call of the Second Age"]				= { ad = -0.10, ind = -0.10 },
	["41140201:Penetrating Cry Attack Speed Buff"]		= { ad = -0.15              },
	["41000D1A:Ripped Throat"]							= {             ind =  0.50 },
--	["41008819:Sign of Power: Command"]					= { ad =  0.15              },
	["411167E1:Sign of Power: Command"]					= { ad =  0.15              },
	["4100881C:Turn up the Heat"]						= { ad =  0.15              },
	["4100881C:Warm Up"]								= {             ind =  0.20 }
};
CCTimerWindow.defaultTimingEffectsLastUpdate = "0.7.1";

-- Colors used.  Colors that are adjacent to each other are assumed to be not easily distinguishable if used next to each other.
-- The colors should form a circle, with the last and first colors being adjacent, as well.
CCTimerWindow.colors = {
	Turbine.UI.Color(1.0, 1.0, 0.0, 0.0),	-- Red
	Turbine.UI.Color(1.0, 1.0, 0.5, 0.0),	-- Orange
	Turbine.UI.Color(1.0, 1.0, 1.0, 0.0),	-- Yellow
	Turbine.UI.Color(1.0, 0.0, 1.0, 0.0),	-- Green
	Turbine.UI.Color(1.0, 0.0, 1.0, 1.0),	-- Cyan
	Turbine.UI.Color(1.0, 0.0, 0.5, 1.0),	-- Light Blue
	Turbine.UI.Color(1.0, 0.0, 0.0, 1.0),	-- Blue
	Turbine.UI.Color(1.0, 0.5, 0.0, 1.0),	-- Purple
	Turbine.UI.Color(1.0, 1.0, 0.0, 1.0)	-- Pink
};

-- Static and default information about the CC skills
-- Defaults: color, duration, apply, start, stop
--   start, stop, and color define the time frame and color used to indicate the "unsafe" time for a skill (i.e, if the skill is used on a Stunned or Dazed target within the time frame, it will hit when the target is Recovering)
--   duration is the duration of the CC effect, and apply is the delay between the skill going on cooldown and the CC effect being applied
-- Statics: act, ind, apply_act, recovery, effect
--   act is the ACT of the skill (the amount of time that the skill spends activating after the induction is finished and before it goes on cooldown)
--   ind is the induction of the skill
--   apply_act relates how changes in the ACT of a skill change the apply delay.  If the ACT of a skill changes by x, then the apply changes by x * apply_act
--   recovery is whether or not the CC effect ends with the Recovering effect
--   effect is the CC effect applied by the skill
CCTimerWindow.ccSkillsStatic = {
	["0x7000B60D"] = { color = 3, duration = 15, apply = -0.172, start = 2.4, stop = 1.3, act = 1.68, ind = 1.0, apply_act = -1.00, recovery =  true, effect = "410D8360:Dazed" },					-- Bane Flare
	["0x70003152"] = { color = 3, duration = 30, apply =  0.000, start = 1.4, stop = 0.3, act = 1.25, ind = 0.0, apply_act = -0.00, recovery =  true, effect = "410D8360:Dazed" },					-- Blinding Flash
	["0x70003143"] = { color = 7, duration = 30, apply = -1.250, start = 4.7, stop = nil, act = 1.50, ind = 2.0, apply_act = -1.00, recovery = false, effect = "410056A2:The Shaking of Arda" },	-- Cracked Earth
	["0x70003EC0"] = { color = 7, duration = 30, apply = -1.250, start = nil, stop = nil, act = 1.50, ind = 0.0, apply_act = -1.00, recovery = false, effect = "4100823D:Rooted" },					-- Herb-lore
	["0x70000FA7"] = { color = 1, duration =  6, apply = -1.547, start = 0.0, stop = 0.0, act = 3.00, ind = 2.0, apply_act = -0.00, recovery =  true, effect = "4100823F:Stunned" },				-- Ents go to War
	["0x70003EC2"] = { color = 1, duration =  3, apply = -1.188, start = 1.2, stop = 0.1, act = 1.50, ind = 1.0, apply_act = -1.00, recovery =  true, effect = "4100823F:Stunned" },				-- Light of the Rising Dawn
	["0x7000B63C"] = { color = 1, duration =  3, apply = -1.234, start = nil, stop = nil, act = 1.50, ind = 0.0, apply_act = -1.00, recovery =  true, effect = "4100823F:Stunned" },				-- Storm-lore
	["0x70003EB9"] = { color = 1, duration =  5, apply = -0.641, start = nil, stop = nil, act = 1.50, ind = 1.5, apply_act = -0.40, recovery =  true, effect = "4100823F:Stunned" }					-- Test of Will
};

-- Default colors for the background and the recovery portions of the display (time 0.0 to -1.0 seconds)
CCTimerWindow.backgroundColorDefault = 4;
CCTimerWindow.recoveryColorDefault = 1;

function CCTimerWindow:Constructor(version)
	Turbine.UI.Window.Constructor(self);

	-- size of a quickslot
	self.size = 36;							

	-- Our display is vertical, we'll use the Y coordinate to indicate how much time is remaining on a CC effect.  Each pixel represents 0.1 seconds.
	-- The pixel representing (self.seconds, self.seconds - 0.1] will be down far enough to fully display a quickslot.
	-- The last pixel in the display will represent (-0.9, -1.0] (0.0 - -1.0 is the Recovering period).
	self.seconds = 30;
	self:SetHeight(self:TimeY(-1.0) + 1);

	self.x = 0;
	self.y = 0;
	self.hide = false;
	self.dragging = false;
	self.move = false;

	self.version = version;
	self.config = { };
	self.curConfig = "";

	self.ccSkills = { };
	for shortcut in pairs(self.ccSkillsStatic) do
		self.ccSkills[shortcut] = { };
	end

	self:LoadSettings();

	-- To represent the display, we use regimes (see CCTimerRegime.lua).
	-- The entire display is a RegimeContainer, while the unsafe zones of each skill are represented using a regime.
	-- The regime container handles display the regimes and their overlapping regions.
	self.regimeContainers = { };
	for _, skill in pairs(self.ccSkills) do
		skill.regimes = { };
	end
	self.regimesRecovery = { };

	self.options = CCTimerOptions(self);

	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.name = self.player:GetName();
	self.tskills = self.player:GetTrainedSkills();
	self.effects = self.player:GetEffects();

	-- CC is the regime container that tracks the durations of CC effects from skills we have used
	self.ccTrack = { };
	self.regimeContainers.cc = CCTimerRegimeContainer(self.colors);
	self.regimeContainers.cc:SetParent(self);
	self.regimeContainers.cc:SetAlpha(0.25);
	self.regimeContainers.cc:SetHeight(self:GetHeight());
	self.regimeContainers.cc:SetPosition(self.size + 4 + 2, 0);
	self.regimeContainers.cc:SetMouseVisible(false);
	for _, skill in pairs(self.ccSkills) do
		skill.regimes.cc = self.regimeContainers.cc:AddRegime();
	end
	self.regimesRecovery.cc = self.regimeContainers.cc:AddRegime();
	self.regimesRecovery.cc:SetBounds(self:TimeY(-0.01), self:TimeY(-1.0));
	self.regimesRecovery.cc:SetVisible(true);

	-- target is the regime container that tracks CC (and other) effects on the current target
	self.regimeContainers.target = CCTimerRegimeContainer(self.colors);
	self.regimeContainers.target:SetParent(self);
	self.regimeContainers.target:SetAlpha(0.25);
	self.regimeContainers.target:SetSize(self.size + 4, self:GetHeight());
	self.regimeContainers.target:SetPosition(0, 0);
	self.regimeContainers.target:SetMouseVisible(false);
	for _, skill in pairs(self.ccSkills) do
		skill.regimes.target = self.regimeContainers.target:AddRegime();
	end
	self.regimesRecovery.target = self.regimeContainers.target:AddRegime();
	self.regimesRecovery.target:SetBounds(self:TimeY(-0.01), self:TimeY(-1.0));
	self.regimesRecovery.target:SetVisible(true);
	self.targetTrack = { };
	self.targetCC = self:NewCCElement(self.regimeContainers.target);

--[[
	-- Tells my target vitals display to ignore effects that we'll be displaying
	Turbine.PluginManager.LoadPlugin("VitalsRacer");
	if _G.vitalsRacerWindow ~= nil then
		self.x = Turbine.UI.Display.GetWidth() * 2 / 3;
		self.y = _G.vitalsRacerWindow:GetTop() - self:GetHeight();
		for effect in pairs(self.ccEffects) do
			_G.vitalsRacerWindow.targetFrame:IgnoreEffect(effect);
		end
		for effect in pairs(self.otherEffects) do
			_G.vitalsRacerWindow.targetFrame:IgnoreEffect(effect);
		end
	end
--]]

	-- track effects on our target that are either CC effects or CC related effects (cracked earth and stun immunity)
	self.fnEffectAdded = function(sender, args)
		local e = self.targetEffects:Get(args.Index);
		local name = e:GetName();
		local icon = e:GetIcon();
		local iconName = string.format("%8X", icon) .. ":" .. name;

		if self.time ~= nil then
			for _, skill in pairs(self.ccSkillsStatic) do
				if skill.effect == iconName then
					self.time.cc = e:GetStartTime();
					self:Calibrate();
					break;
				end
			end
		end

		if self.ccEffects[iconName] ~= nil then
			self.targetCCName = iconName;
			self.targetCCID = e:GetID();
			if self.ccEffects[iconName] ~= 0 then
				self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, self.ccEffects[iconName]));
			end
			self.targetCC.startTime = e:GetStartTime();
			self.targetCC.endTime = self.targetCC.startTime + e:GetDuration();
			self.targetCC.control:SetVisible(true);
			self:UpdateLayout();
		elseif self.otherEffects[iconName] ~= nil then
			local item = self:NewCCElement(self.regimeContainers.target);
			item.type = self.otherEffects[iconName];
			item.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, item.type));
			item.startTime = e:GetStartTime();
			item.duration = e:GetDuration();
			item.endTime = item.startTime + e:GetDuration();
			item.ID = e:GetID();
			item.control:SetVisible(true);
			table.insert(self.targetTrack, item);
			self:UpdateLayout();
		end
	end;
	self.fnEffectRemoved = function(sender, args)
		local id = args.Effect:GetID();

		if id == self.targetCCID then
			self.targetCCName = nil;
			self.targetCCID = nil;
			self.targetCC.control:SetVisible(false);
			self:UpdateLayout();
		else
			for i, effect in ipairs(self.targetTrack) do
				if effect.ID == id then
					self:DeleteCCElement(effect);
					table.remove(self.targetTrack, i);
					self:UpdateLayout();
					break;
				end
			end
		end
	end;
	self.fnEffectsCleared = function(sender, args)
		if self.targetCCName ~= nil or #self.targetTrack > 0 then
			self.targetCCName = nil;
			self.targetCCID = nil;
			self.targetCC.control:SetVisible(false);

			for _, effect in ipairs(self.targetTrack) do
				self:DeleteCCElement(effect);
			end
			self.targetTrack = { };

			self:UpdateLayout();
		end
	end;

	-- Hook our CC skills.  We'll need to know when they've been activated (gone on cooldown)
	local fResetTimeChanged = function(sender, reset) self:SkillUse(sender:GetShortcut()); end;
	local fActiveChanged = function(sender, active)
		local skill = self.ccSkills[sender:GetShortcut()];
		skill.regimes.cc:SetVisible(active);
		skill.regimes.target:SetVisible(active);
		skill.regimes.options:SetVisible(active);
	end;
	for shortcut, skill in pairs(self.ccSkills) do
		skill.activeSkill = moebius92.ActiveSkills.ActiveSkill(shortcut);
		skill.activeSkill.ResetTimeChanged = fResetTimeChanged;
		skill.activeSkill.ActiveChanged = fActiveChanged;
		local active = skill.activeSkill:IsActive();
		skill.regimes.cc:SetVisible(active);
		skill.regimes.target:SetVisible(active);
		skill.regimes.options:SetVisible(active);
	end

	self.m92TrainedSkills = moebius92.ActiveSkills.TrainedSkills();
	self.m92TrainedSkills.SkillsChanged = function(sender, state) self:AutoLoadConfig(state) end;
	self.m92TrainedSkills.current = "";
	self:UpdateConfigShortcuts();
	moebius92.Handlers.AddChatHandler(function(m) self:ChatMessage(m) end, Turbine.ChatType.PlayerCombat, self.name);
	moebius92.Handlers.AddEventHandler(self.player, "TargetChanged", function() self:StopCalibrate(); self:UpdateTarget(); end);

	-- Track effects on ourself that modify our induction times or our attack duration.  Also track the Feral buff that increases stun durations
	self.curTimingEffects = { };
	self.stunDuration = 0;

	moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded",
function(sender, args)
	local e = self.effects:Get(args.Index);
	local name = e:GetName();
	local icon = e:GetIcon();
	local iconName = string.format("%8X", icon) .. ":" .. name;

	if iconName == "41139EF1:Feral - Bog-guardian" then
		self.stunDuration = self.stunDuration + 2;
	elseif self.timingEffects[iconName] ~= nil then
		local duration = e:GetDuration();

		-- Handle the Enfeeble trait
		if duration == 10 and (iconName == "41008819:Sign of Power: Command" or iconName == "411167E1:Sign of Power: Command") then
			table.insert(self.curTimingEffects, { id = e:GetID(), ind = (self.timingEffects[iconName].ind or 0.0) * 2, ad = (self.timingEffects[iconName].ad or 0.0) * 2, endTime = e:GetStartTime() + e:GetDuration() });
		else
			table.insert(self.curTimingEffects, { id = e:GetID(), ind = self.timingEffects[iconName].ind or 0.0, ad = self.timingEffects[iconName].ad or 0.0, endTime = e:GetStartTime() + e:GetDuration() });
		end
		table.sort(self.curTimingEffects, function(a, b) return a.endTime > b.endTime; end);
		self:UpdateRegimes();
	end
	self.options:LogEffect(icon, name);
end);
	moebius92.Handlers.AddEventHandler(self.effects, "EffectRemoved",
function(sender, args)
	local e = args.Effect;
	local name = e:GetName();
	local icon = e:GetIcon();
	local iconName = string.format("%8X", icon) .. ":" .. name;

	if iconName == "41139EF1:Feral - Bog-guardian" then
		self.stunDuration = self.stunDuration - 2;
	elseif self.timingEffects[iconName] ~= nil then
		local id = e:GetID();

		for i, effect in ipairs(self.curTimingEffects) do
			if effect.id == id then
				table.remove(self.curTimingEffects, i);
				break;
			end
		end
		self:UpdateRegimes();
	end
end);
	moebius92.Handlers.AddEventHandler(self.effects, "EffectsCleared", function() self.stunDuration = 0; self.timingEffects = { }; self:UpdateRegimes(); end);

	for i = 1, self.effects:GetCount(), 1 do
		local e = self.effects:Get(i);
		local name = e:GetName();
		local icon = e:GetIcon();
		local iconName = string.format("%8X", icon) .. ":" .. name;

		if iconName == "41139EF1:Feral - Bog-guardian" then
			self.stunDuration = self.stunDuration + 2;
		elseif self.timingEffects[iconName] ~= nil then
			table.insert(self.curTimingEffects, { id = e:GetID(), ind = self.timingEffects[iconName].ind or 0.0, ad = self.timingEffects[iconName].ad or 0.0, endTime = e:GetStartTime() + e:GetDuration() });
		end
	end
	table.sort(self.curTimingEffects, function(a, b) return a.endTime > b.endTime; end);

	self:SetPosition(self.x, self.y);
	self:SetWantsKeyEvents(true);
	self:SetMouseVisible(false);
	self:SetVisible(true);
	self:UpdateRegimes();
	self:UpdateLayout();
	self:UpdateTarget();
end

-- Given a time, calculate the corresponding Y position
function CCTimerWindow:TimeY(t)
	if t == nil then
		return nil;
	end
	return self.size + self.seconds * 10 - math.floor(t * 10);
end

-- Calculate the ACT of a skill, taking into account attack duration (bonus)
function CCTimerWindow:GetAct(shortcut, bonus)
	local act = self.ccSkillsStatic[shortcut].act;

	bonus = math.min(0.40, bonus);
	bonus = math.max(-1.00, bonus);

	if bonus < 0.00 then
		act = act * (1 + bonus / 5.0);
	elseif bonus > 0.00 then
		act = act * (1 + bonus / 3.5);
	end

	return act;
end

-- calculate the induction of a skill, taking into account some % modifier (bonus)
function CCTimerWindow:GetInduction(shortcut, bonus)
	bonus = math.max(-1.00, bonus);
	return self.ccSkillsStatic[shortcut].ind * (1.00 + bonus);
end

-- Generate a quickslot to be placed on the display (represents some sort of CC/CC related effect)
function CCTimerWindow:NewCCElement(parent)
	local element = { };

	element.control = Turbine.UI.Control();
	element.control:SetSize(self.size, self.size);
	element.control:SetZOrder(1);
	element.control:SetVisible(false);
	element.control:SetMouseVisible(false);
	element.control:SetParent(parent);
	element.quickslot = Turbine.UI.Lotro.Quickslot();
	element.quickslot:SetParent(element.control);
	element.quickslot:SetPosition(-1, -1);
	element.quickslot:SetSize(self.size, self.size);
	element.quickslot:SetMouseVisible(false);
	return element;
end

function CCTimerWindow:DeleteCCElement(element)
	element.control:SetParent(nil);
	element.control:SetVisible(false);
end

-- Target changed, requires cleaning up old handlers and target display, hooking new handlers, populating target display
function CCTimerWindow:UpdateTarget()
	if self.target ~= nil then
		for event, handler in pairs(self.targetEffectsHandlers) do
			moebius92.Handlers.RemoveEventHandler(self.targetEffects, event, handler);
		end
	end

	self.target = nil;
	self.targetEffects = nil;
	self.targetEffectsHandlers = nil;
	self.targetCCName = nil;
	self.targetCCID = nil;
	self.targetCC.control:SetVisible(false);

	for _, effect in ipairs(self.targetTrack) do
		self:DeleteCCElement(effect);
	end
	self.targetTrack = { };

	self.target = self.player:GetTarget();
	if self.target ~= nil and self.target:IsA(Turbine.Gameplay.Actor) then
		self.targetEffects = self.target:GetEffects();
		self.targetEffectsHandlers = { };
		self.targetEffectsHandlers.EffectAdded = moebius92.Handlers.AddEventHandler(self.targetEffects, "EffectAdded", self.fnEffectAdded);
		self.targetEffectsHandlers.EffectRemoved = moebius92.Handlers.AddEventHandler(self.targetEffects, "EffectRemoved", self.fnEffectRemoved);
		self.targetEffectsHandlers.EffectsCleared = moebius92.Handlers.AddEventHandler(self.targetEffects, "EffectsCleared", self.fnEffectsCleared);

		for i = 1, self.targetEffects:GetCount(), 1 do
			local e = self.targetEffects:Get(i);
			local name = e:GetName();
			local icon = e:GetIcon();
			local iconName = string.format("%8X", icon) .. ":" .. name;

			if self.ccEffects[iconName] ~= nil then
				self.targetCCName = iconName;
				self.targetCCID = e:GetID();
				if self.ccEffects[iconName] ~= 0 then
					self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, self.ccEffects[iconName]));
				else
					self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, "0x70003F0D"));
				end
				self.targetCC.startTime = e:GetStartTime();
				self.targetCC.endTime = self.targetCC.startTime + e:GetDuration();
				self.targetCC.control:SetVisible(true);
			elseif self.otherEffects[iconName] ~= nil then
				local item = self:NewCCElement(self.regimeContainers.target);
				item.type = self.otherEffects[iconName];
				item.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, item.type));
				item.startTime = e:GetStartTime();
				item.duration = e:GetDuration();
				item.endTime = item.startTime + e:GetDuration();
				item.ID = e:GetID();
				item.control:SetVisible(true);
				table.insert(self.targetTrack, item);
			end
		end
	else
		self.target = nil;
	end

	self:UpdateLayout();
end

function CCTimerWindow:UpdateConfigShortcuts()
	if self.m92TrainedSkills ~= nil then
		local shortcuts = { };
		for _, config in pairs(self.config) do
			if config.skill ~= nil then
				table.insert(shortcuts, config.skill);
			end
		end

		table.sort(shortcuts);
		local serial = table.concat(shortcuts, ",");

		if serial ~= self.m92TrainedSkills.current then
			self.m92TrainedSkills.current = serial;
			self.m92TrainedSkills:SetShortcuts(shortcuts);
		end
	end
end

function CCTimerWindow:AutoLoadConfig(state)
	-- configs holds the set of possible configurations, based on the currently active/available skills
	local configs = { };
	for label, config in pairs(self.config) do
		if config.skill ~= nil and state[config.skill] then
			table.insert(configs, label);
		end
	end

	if #configs == 1 then
		if configs[1] ~= self.curConfig then
			self:LoadConfig(configs[1]);
			return;
		end
	elseif #configs > 1 then
		table.sort(configs);
		Turbine.Shell.WriteLine("Multiple possible configs: " .. table.concat(configs, ", "));
	end
	self.options:UpdateEntries();
end

-- On skill use, record the time the skill was used and try to match it with a hit message from the combat log
function CCTimerWindow:SkillUse(shortcut)
	local time = Turbine.Engine.GetGameTime();
	local cooldown = self.ccSkills[shortcut].activeSkill:GetCooldown();
	local rtime = self.ccSkills[shortcut].activeSkill:GetResetTime();

	if rtime >= 0 and (math.abs(time - (rtime - cooldown)) < 0.1) then
		self.ccSkills[shortcut].useTime = rtime - cooldown;
		if self.ccSkills[shortcut].hitTime ~= nil and (self.ccSkills[shortcut].useTime - self.ccSkills[shortcut].hitTime) < 1.5 then
			if self.time ~= nil then
				self.time.reset = self.ccSkills[shortcut].useTime;
				self:Calibrate();
			end
			self:AddCC(shortcut, self.ccSkills[shortcut].useTime);
			self.ccSkills[shortcut].useTime = nil;
			self.ccSkills[shortcut].hitTime = nil;
		end
	end

	self:UpdateRegimes();
end

-- Look for hit messages in the combat log.  Try to match them with skill uses.
function CCTimerWindow:ChatMessage(m)
	m = string.gsub(m, "<[^>]*>", "");
	local i = string.find(m, " ");

	if i ~= nil then
		local name = trim(string.sub(m, 1, i - 1));
		local rest = trim(string.sub(m, i + 1));
		if name == self.name then
			if string.sub(rest, 1, 9) == "scored a " then
				local j = string.find(rest, " with ");
				i = string.find(rest, " on ");
				if i ~= nil then
					local skillName = string.sub(rest, j + 6, i - 1);
					if string.sub(skillName, 1, 9) == "Improved " then
						skillName = string.sub(skillName, 10);
					end
					if string.sub(skillName, 1, 9) == "Advanced " then
						skillName = string.sub(skillName, 10);
					end

					for shortcut, skill in pairs(self.ccSkills) do
						if moebius92.SkillData.shortcut[shortcut] == skillName then
							skill.hitTime = Turbine.Engine.GetGameTime();
							if skill.useTime ~= nil and (skill.hitTime - skill.useTime) < 1.5 then
								if self.time ~= nil then
									self.time.reset = skill.useTime;
									self:Calibrate();
								end
								self:AddCC(shortcut, skill.useTime);
								skill.useTime = nil;
								skill.hitTime = nil;
							end
						end
					end
				end
			end
		end
	end
end

-- A CC skill has been used and hit something.  Now we need to track it
function CCTimerWindow:AddCC(shortcut, useTime)
	local cc = self:NewCCElement(self.regimeContainers.cc);
	cc.type = shortcut;

	-- Take into account how attack duration bonuses may have changed the skill's apply delay
	if self.config[self.curConfig].attdur ~= nil then
		local bonusAD = 0.0;
		for _, effect in ipairs(self.curTimingEffects) do
			bonusAD = bonusAD + effect.ad;
		end

		local actDelta = self:GetAct(shortcut, self.config[self.curConfig].attdur + bonusAD) - self:GetAct(shortcut, self.config[self.curConfig].attdur);

		cc.startTime = useTime + self.config[self.curConfig][shortcut].apply + self.ccSkillsStatic[shortcut].apply_act * actDelta;
	else
		cc.startTime = useTime + self.config[self.curConfig][shortcut].apply;
	end

	cc.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, shortcut));
	if self.config[self.curConfig][shortcut].preroot ~= nil then
		cc.endTime = cc.startTime + self.config[self.curConfig][shortcut].preroot;
		cc.pre = true;
	else
		if self.ccSkillsStatic[shortcut].effect == "4100823F:Stunned" then
			cc.endTime = cc.startTime + self.config[self.curConfig][shortcut].duration + self.stunDuration;
		else
			cc.endTime = cc.startTime + self.config[self.curConfig][shortcut].duration;
		end
		cc.pre = false;
	end
	table.insert(self.ccTrack, cc);

	self:AssignLanes();
end

-- Assign a lane for each CC effect we're tracking
function CCTimerWindow:AssignLanes()
	local time = Turbine.Engine.GetGameTime();

	for _, cc in ipairs(self.ccTrack) do
		if cc.startTime > time then
			cc.top = nil;
			cc.bottom = nil;
		else
			cc.bottom = (cc.endTime - time) * 10;
			cc.top = cc.bottom + self.size;
		end
	end
	table.sort(self.ccTrack, function(a, b) return (a.top or math.huge) < (b.top or math.huge); end);
	moebius92.RaceTrack.layout(self.ccTrack);

	self:UpdateLayout();
end

-- Determine if the target and CC displays should be visible, determine the width of the CC display
function CCTimerWindow:UpdateLayout()
	for _, cc in ipairs(self.ccTrack) do
		if cc.lane == nil then
			cc.control:SetVisible(false);
		else
			cc.control:SetVisible(true);
		end
	end

	if self.targetCCName == "410E7686:Recovering" then
		for _, cc in ipairs(self.ccTrack) do
			if self.ccSkillsStatic[cc.type].recovery and math.abs(self.targetCC.endTime - (cc.endTime + 1.0)) < 0.1 then
				self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, cc.type));
				cc.control:SetVisible(false);
				break;
			end
		end
	elseif self.targetCCName ~= nil then
		for _, cc in ipairs(self.ccTrack) do
			if self.ccSkillsStatic[cc.type].effect == self.targetCCName and math.abs(self.targetCC.endTime - cc.endTime) < 0.1 then
				self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, cc.type));
				cc.control:SetVisible(false);
				cc.endTime = self.targetCC.endTime
				break;
			end
		end
	end

	-- Determine if a CC effect we're tracking is on the current target
	for _, effect in ipairs(self.targetTrack) do
		for _, cc in ipairs(self.ccTrack) do
			if effect.type == cc.type and math.abs(effect.endTime - cc.endTime) < 0.1 then
				cc.control:SetVisible(false);
			end
		end
	end

	local w = 0;
	for _, cc in ipairs(self.ccTrack) do
		if cc.control:IsVisible() and cc.lane > w then
			w = cc.lane;
		end
	end

	if w == 0 then
		self:SetWidth(self.size + 4);
		self.regimeContainers.cc:SetVisible(false);
	else
		self:SetWidth(self.size + 4 + 2 + self.size * w + 4);
		self.regimeContainers.cc:SetWidth(self.size * w + 4);
		self.regimeContainers.cc:SetVisible(true);
	end

	if self:IsMouseVisible() or self.targetCCName ~= nil or #self.targetTrack > 0 or (self.regimeContainers.cc:IsVisible() and self.target ~= nil) then
		self.regimeContainers.target:SetVisible(true);
	else
		self.regimeContainers.target:SetVisible(false);
	end

	self:Update();
end

-- Move the effects that we're tracking.  Remove them from the display if they're out of time
function CCTimerWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local changed = false;
	local removeIndices = { };

	for i, cc in ipairs(self.ccTrack) do
		local delta = cc.endTime - time;

		if delta < -1.0 or (delta < 0.0 and not self.ccSkillsStatic[cc.type].recovery) then
			changed = true;
			if cc.pre then
				cc.pre = false;
				cc.endTime = cc.endTime + self.config[self.curConfig][cc.type].duration;
				cc.type = "0x70003EC0";
				cc.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, "0x70003EC0"));
			else
				self:DeleteCCElement(cc);
				table.insert(removeIndices, i);
			end
		elseif cc.startTime <= time and cc.lane == nil then
			changed = true;
		end
	end

	table.sort(removeIndices, function(a, b) return a > b; end);
	for _, i in ipairs(removeIndices) do
		table.remove(self.ccTrack, i);
	end

	if changed then
		self:AssignLanes();
		return;
	end

	removeIndices = { };
	changed = false;

	for i, effect in ipairs(self.targetTrack) do
		local delta = effect.endTime - time;
		if delta < 0.0 then
			self:DeleteCCElement(effect);
			table.insert(removeIndices, i);
			changed = true;
		end
	end

	table.sort(removeIndices, function(a, b) return a > b; end);
	for _, i in ipairs(removeIndices) do
		table.remove(self.targetTrack, i);
	end

	if changed then
		self:UpdateLayout();
		return;
	end

	local update = false;
	for _, cc in ipairs(self.ccTrack) do
		if cc.control:IsVisible() then
			local y = self:TimeY(cc.endTime - time);
			cc.control:SetPosition(self.size * (cc.lane - 1) + 2, y - self.size + 1);
			cc.control:SetBackColor(self.regimeContainers.cc:GetColor(y));
			update = true;
		end
	end

	if self.targetCCName ~= nil then
		local t = self.targetCC.endTime - time;
		if t <= 0.0 then
			self.targetCCName = nil;
			self.targetCCID = nil;
			self.targetCC.control:SetVisible(false);
			self:UpdateLayout();
		elseif self.targetCCName ~= "410E7686:Recovering" then
			local y = self:TimeY(t);
			self.targetCC.control:SetPosition(2, y - self.size + 1);
			self.targetCC.control:SetBackColor(self.regimeContainers.cc:GetColor(y));
		else
			local y = self:TimeY(t - 1.0);
			self.targetCC.control:SetPosition(2, y - self.size + 1);
			self.targetCC.control:SetBackColor(self.regimeContainers.cc:GetColor(y));
		end
		update = true;
	end

	for _, effect in ipairs(self.targetTrack) do
		if effect.duration > 1000000 then
			local y = self:TimeY(self.seconds);
			effect.control:SetPosition(2, y - self.size + 1);
			effect.control:SetBackColor(self.regimeContainers.cc:GetColor(y));
		else
			local y = self:TimeY(effect.endTime - time);
			effect.control:SetPosition(2, y - self.size + 1);
			effect.control:SetBackColor(self.regimeContainers.cc:GetColor(y));
			update = true;
		end
	end

	self:SetWantsUpdates(update);
end

function CCTimerWindow:LoadConfig(config)
	if self.config[config] == nil then
		Turbine.Shell.WriteLine("Config " .. self.curConfig .. " not found.");
	else
		self.curConfig = config;
		self:UpdateRegimes();
		self:SaveSettings();
		self.options:UpdateEntries();
		Turbine.Shell.WriteLine("Config " .. self.curConfig .. " loaded.");
	end
end

function CCTimerWindow:GetConfigs()
	local configs = { };

	for label in pairs(self.config) do
		table.insert(configs, label);
	end

	table.sort(configs);
	return configs;
end

-- Update the skill unsafe zone regimes, based on the current configuration
function CCTimerWindow:UpdateRegimes()
	for _, regime in pairs(self.regimeContainers) do
		regime:SetColor(self.config[self.curConfig].back.color);
	end
	for _, regime in pairs(self.regimesRecovery) do
		regime:SetColor(self.config[self.curConfig].rec.color);
	end

	local config = self.config[self.curConfig];
	for shortcut, skill in pairs(self.ccSkills) do
		local start = config[shortcut].start;
		local stop = config[shortcut].stop;

		if skill.activeSkill:IsActive() then
			local resetTime = skill.activeSkill:GetResetTime();
			local bonusInd = 0.0;
			local bonusAD = 0.0;

			for _, effect in ipairs(self.curTimingEffects) do
				if effect.endTime < resetTime then
					break;
				end

				bonusInd = bonusInd + effect.ind;
				bonusAD = bonusAD + effect.ad;
			end

			if config.induction ~= nil then
				local indDelta = self:GetInduction(shortcut, config.induction + bonusInd) - self:GetInduction(shortcut, config.induction);
				if start ~= nil then start = start + indDelta; end
				if stop ~= nil then stop = stop + indDelta; end
			end
			if config.attdur ~= nil then
				local actDelta = self:GetAct(shortcut, config.attdur + bonusAD) - self:GetAct(shortcut, config.attdur);
				if start ~= nil then start = start + actDelta; end
				if stop ~= nil then stop = stop + actDelta; end
			end

		end
		local startYreal = self:TimeY(start);
		local stopYreal = self:TimeY(stop);
		if startYreal ~= nil and (self.ccSkillsStatic[shortcut].effect == "4100823D:Rooted" or self.ccSkillsStatic[shortcut].effect == "410056A2:The Shaking of Arda") then
			startYreal = startYreal + 1;
		end

		start = config[shortcut].start;
		stop = config[shortcut].stop;
		if config.induction ~= nil and self.options.induction ~= nil then
			local indDelta = self:GetInduction(shortcut, config.induction + self.options.induction) - self:GetInduction(shortcut, config.induction);
			if start ~= nil then start = start + indDelta; end
			if stop ~= nil then stop = stop + indDelta; end
		end
		if config.attdur ~= nil and self.options.attdur ~= nil then
			local actDelta = self:GetAct(shortcut, config.attdur + self.options.attdur) - self:GetAct(shortcut, config.attdur);
			if start ~= nil then start = start + actDelta; end
			if stop ~= nil then stop = stop + actDelta; end
		end
		local startYpreview = self:TimeY(start);
		local stopYpreview = self:TimeY(stop);

		for label, regime in pairs(skill.regimes) do
			regime:SetColor(config[shortcut].color);
			if label == "cc" or label == "target" then
				regime:SetBounds(startYreal, stopYreal);
			else
				regime:SetBounds(startYpreview, stopYpreview);
			end
		end
	end
end

-- Calibrate pretty much runs off of the CC tracking.  Also keeps track of when a quickslot activating key was pressed
function CCTimerWindow:ToggleCalibrate()
	if self.time == nil then
		self:StartCalibrate();
	else
		self:StopCalibrate();
	end
end

function CCTimerWindow:StartCalibrate()
	self.time = { };
	self.options:StartCalibrate();
end

function CCTimerWindow:Calibrate()
	if self.time.cc ~= nil and self.time.reset ~= nil then
		local apply = self.time.cc - self.time.reset;
		local key;

		Turbine.Shell.WriteLine(string.format("Apply: %.3f", apply));
		if self.time.key ~= nil then
			key = self.time.cc - self.time.key;
			Turbine.Shell.WriteLine(string.format("Key: %.3f", key));
		end
		self.options:Calibrate(apply, key);
		self:StopCalibrate();
	end
end

function CCTimerWindow:StopCalibrate()
	self.time = nil;
	self.options:StopCalibrate();
end

function CCTimerWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function CCTimerWindow:MouseMove(args)
	if self.dragging then
		local x, y = self:GetPosition();
		x = x + (args.X - self.dragStartX);
		y = y + (args.Y - self.dragStartY);
		self.x = x;
		self.y = y;
		self:SetPosition(x, y);
	end
end

function CCTimerWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self:SaveSettings();
		self.dragging = false;
	end
end

function CCTimerWindow:KeyDown(args)
	if args.Action == 0x1000007B then
		if self.dragging then
			self:SaveSettings();
			self.dragging = false;
		end
		self:SetMouseVisible(not self:IsMouseVisible());
		self:UpdateLayout();
	elseif args.Action == 0x100000B3 then
		self:SetVisible(not self:IsVisible());
	elseif self.keymap[args.Action] ~= nil and self.time ~= nil then
		self.time.key = Turbine.Engine.GetGameTime();
	end
end

function CCTimerWindow:SaveSettings()
	local t = { };

	self:UpdateConfigShortcuts();

	t.version = self.version;

	t.x = self.x;
	t.y = self.y;
	t.config = self.config;
	t.curConfig = self.curConfig;
	t.timingEffects = self.timingEffects;

	for _, config in pairs(t.config) do
		for _, element in pairs(config) do
			if type(element) == "table" then
				element.r = self.colors[element.color].R;
				element.g = self.colors[element.color].G;
				element.b = self.colors[element.color].B;
			end
		end
	end

	Turbine.PluginData.Save(Turbine.DataScope.Character, "CCTimer", serialize(t));
end

function CCTimerWindow:AddDefaultTimingEffects()
	local added = { };

	for iconName, timing in pairs(self.defaultTimingEffects) do
		if self.timingEffects[iconName] == nil then
			table.insert(added, iconName);
			self.timingEffects[iconName] = { };
			for k, v in pairs(timing) do
				self.timingEffects[iconName][k] = v;
			end
		end
	end

	return added;
end

function CCTimerWindow:LoadSettings()
	local t = Turbine.PluginData.Load(Turbine.DataScope.Character, "CCTimer");
	if type(t) == "string" then
		t = loadstring("return " .. t .. ";")();
	end

	if type(t) == "table" then
		if versioncmp(t.version, "0.6") >= 0 then
			self.x = t.x;
			self.y = t.y;
			self.config = t.config;
			self.curConfig = t.curConfig;

			self.timingEffects = { };
			if t.timingEffects ~= nil then
				self.timingEffects = t.timingEffects;
			end

			if versioncmp(t.version, self.defaultTimingEffectsLastUpdate) < 0 then
				self:AddDefaultTimingEffects();
			end

			if versioncmp(t.version, "0.6") == 0 then
				for _, config in pairs(self.config) do
					for _, element in pairs(config) do
						if type(element) == "table" then
							element.apply = element.act;
							element.act = nil;
						end
					end
				end
			end

			for _, config in pairs(self.config) do
				for _, element in pairs(config) do
					if type(element) == "table" then
						local distBest;
						local colorBest;

						for color, tc in ipairs(self.colors) do
							local dist = math.abs(element.r - tc.R) + math.abs(element.g - tc.G) + math.abs(element.b - tc.B);
							if distBest == nil or dist < distBest then
								distBest = dist;
								colorBest = color;
							end
						end

						element.color = colorBest;
						element.r = nil;
						element.g = nil;
						element.b = nil;
					end
				end
			end
		end
	end

	if self.config[self.curConfig] == nil then
		self.curConfig = "default";
		self.config[self.curConfig] = { };
		for shortcut, skill in pairs(self.ccSkillsStatic) do
			self.config[self.curConfig][shortcut] = { };
			self.config[self.curConfig][shortcut].duration = skill.duration;
			self.config[self.curConfig][shortcut].apply = skill.apply;
			self.config[self.curConfig][shortcut].start = skill.start;
			self.config[self.curConfig][shortcut].stop = skill.stop;
			self.config[self.curConfig][shortcut].color = skill.color;

			if self.ccEffects[skill.effect] == nil then
				self.config[self.curConfig][shortcut].preroot = 3;
			end
		end

		self.config[self.curConfig].back = { color = self.backgroundColorDefault };
		self.config[self.curConfig].rec = { color = self.recoveryColorDefault };
	end
end